import json

from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse, QueryDict

# todo: try for import config
from .utils.data_functions import save_traceback
from avishan_config import USER_RESPONSE_DICT_FIELDS, MONITORED_URLS_START, NOT_MONITORED_URLS, \
    MONITORED_ANONYMOUS_URLS
from .exceptions import AuthException
from .models import UserUserGroup
from .utils.auth_functions import decode_token, verify_user, generate_token
from avishan_wrapper import current_request


# todo: disable unused middlewares and installed apps
# todo: bayad ye chizi dashte bashim k startup command bashe. role haro doros kone o yeseri kar bara re run
# todo readme ro doros konam
# todo: age ba ye esm, 2ta model dashtim che konim?
# todo: csrf chie? az middleware hazf she ya doros she?
# todo: 299 for message to user
# todo add username/password enter mode
# todo add email/otp enter mode
# todo have instant turn off exception record button

class AvishanMiddleware:
    def __init__(self, get_response):

        self.get_response = get_response
        from .utils.documentation_functions import create_raml_documentation
        # create_raml_documentation(PROJECT_NAME) todo: aya harbar bayad ejra she?

    def __call__(self, request: WSGIRequest):
        try:
            if must_have_token(request.path):
                try:
                    self.token = request.META['HTTP_TOKEN']
                except KeyError as e:
                    save_traceback()
                    raise AuthException(AuthException.TOKEN_ERROR)
                self.decoded_token = decode_token(self.token)
                try:
                    user_user_group = UserUserGroup.get(avishan_raise_exception=True,
                                                        id=self.decoded_token['user_user_group_id'])
                except UserUserGroup.DoesNotExist:
                    save_traceback()
                    raise AuthException(AuthException.ACCOUNT_NOT_FOUND)
                current_request['user'] = user_user_group.user

                # prints user data to log
                if user_user_group.user.have_profile:
                    print(
                        user_user_group.user.phone + " " + user_user_group.user.first_name + " " + user_user_group.user.last_name)
                else:
                    print(user_user_group.user.phone)

                current_request['user_group'] = user_user_group.user_group
                verify_user()

            self.data_fields_on_request(request)

            response = self.get_response(request)
            if response.status_code == 567:
                return JsonResponse({}, status=567)
            # todo inja k miad "status_code" 500 age bood male khodemoon nis dige
            # todo check response to to_dict models

            if isinstance(response, JsonResponse):
                # todo: merge response dict + current_request.response
                status_code = response.status_code
                if must_have_token(request.path):
                    current_request['response']['token'] = generate_token(self.decoded_token, self.token,
                                                                          current_request['user'],
                                                                          current_request['user_group'])
                    current_request['response']['user'] = current_request['user'].to_dict(
                        visible_list=USER_RESPONSE_DICT_FIELDS)
                    if current_request['execution_time'] > 0:
                        print("EXEC:", current_request['execution_time'])
                        current_request['response']['execution_time'] = current_request['execution_time']

                    return JsonResponse(current_request['response'], status=status_code)

                elif current_request['execution_time'] > 0:
                    print("EXEC:", current_request['execution_time'])
                    response = json.loads(response.content.decode('utf-8'))
                    response['execution_time'] = current_request['execution_time']
                    # clear_current_request(current_request)
                    return JsonResponse(response, status=status_code)
            # clear_current_request(current_request)
            return response
        except Exception as e:
            save_traceback()
            current_request['exception'] = e
            return JsonResponse({}, status=567)

    @staticmethod
    def data_fields_on_request(request: WSGIRequest):
        if request.method == 'GET':
            params = request.GET
        elif request.method == 'POST':
            params = request.POST
        elif request.method == 'PUT' or request.method == 'DELETE':
            params = QueryDict(request.body)

        request.search = params.get('search')
        request.page = params.get('page')
        request.page_size = params.get('page_size')
        request.filter = params.get('filter')
        request.sort = params.get('sort')

        if not request.search:
            request.search = ''
        if not request.page:
            request.page = 0
        if not request.page_size:
            request.page_size = 20
        if not request.filter:
            request.filter = []
        if not request.sort:
            request.sort = []


def must_monitor(url: str) -> bool:
    for start in MONITORED_URLS_START:
        if url.startswith(start):
            for not_monitored_start in NOT_MONITORED_URLS:
                if url.startswith(not_monitored_start):
                    return False
            return True
    return False


def must_have_token(url: str) -> bool:
    if must_monitor(url):
        for start in MONITORED_ANONYMOUS_URLS:
            if url.startswith(start):
                return False
    else:
        return False
    return True



# Generated by Django 2.2.3 on 2019-08-18 03:36

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0004_accessrule_crud_type'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='exceptionrecord',
            name='http_method',
        ),
        migrations.RemoveField(
            model_name='exceptionrecord',
            name='url',
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='datetime',
            field=models.DateTimeField(default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='request_data',
            field=models.TextField(default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='request_method',
            field=models.CharField(default=None, max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='request_url',
            field=models.TextField(default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='response',
            field=models.TextField(default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='status_code',
            field=models.IntegerField(default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='traceback',
            field=models.TextField(default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='avishan.User'),
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='user_group',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='avishan.UserGroup'),
        ),
    ]



import re
from typing import List, Optional, Type, Tuple

from django.db import models

# todo: encrypted fields: https://django-fernet-fields.readthedocs.io/en/latest/
from django.db.models import QuerySet
from django.db.models.fields import Field, NOT_PROVIDED

from .utils.bch_datetime import BchDatetime


# todo: check access rules in crud
# todo: create a helping class to move many of functions there
class AvishanModel(models.Model):
    date_created = models.DateTimeField(auto_now_add=True)

    compact_fields = []
    private_fields = []
    added_properties = []
    # django admin
    date_hierarchy = None
    list_display = None
    list_filter = []
    list_max_show_all = 300
    list_per_page = 100
    raw_id_fields = []
    readonly_fields = ['date_created']
    search_fields = []

    class ObjectAction:
        GET = 'GET'
        CREATE = 'CREATE'
        UPDATE = 'UPDATE'

    # todo: auto add added properties
    # todo: have to_dict property for each method
    # todo: have private methods for main objects crud start with _
    # todo: on update, create cannot change pk
    class Meta:
        abstract = True

    @classmethod
    def get_fields(cls) -> List[Field]:
        return cls._meta.fields

    @classmethod
    def get_full_fields(cls) -> List[Field]:
        return cls.get_fields() + cls._meta.many_to_many

    @staticmethod
    def is_field_required(field: Field) -> bool:
        if field.name == 'id' or field.default != NOT_PROVIDED or \
                (isinstance(field, models.DateField) or isinstance(field, models.TimeField) and (
                        field.auto_now or field.auto_now_add)):
            return False
        if isinstance(field, models.ManyToManyField):
            return False

        if field.blank or field.null:
            return False

        # todo: is it true?
        return True

    @staticmethod
    def is_field_readonly(field: Field) -> bool:
        if (isinstance(field, models.DateField) or isinstance(field, models.DateTimeField) or
            isinstance(field, models.TimeField)) and (field.auto_now or field.auto_now_add):
            return True
        return False

    @classmethod
    def is_field_identifier_for_model(cls, field: Field, on_create: bool = False) -> bool:
        if on_create and isinstance(field, models.OneToOneField):
            return False
        if on_create and not field.primary_key:
            return False
        return field.primary_key or field.unique

    @classmethod
    def get_readonly_fields(cls) -> list:
        return [field for field in cls.get_fields() if cls.is_field_readonly(field)]

    def __str__(self):
        for field in self.get_fields():
            if isinstance(field, models.CharField):
                return self.__getattribute__(field.name)
        return super().__str__()
        # todo return first field str

    @classmethod
    def get(cls, avishan_raise_400: bool = False,
            avishan_raise_exception: bool = False,
            object_trace: str = "", **kwargs) -> \
            Optional['AvishanModel']:
        # todo oon halati k query mizanim ba __ ha ham bayad vojood dashte bashe
        """
        [X] if not found return None
        [X] if bch datetime sent, convert to datetime
        """
        # todo: remove non identifier attributes to prevent multi level GET

        try:
            return cls.__get(**kwargs)
        except cls.DoesNotExist as e:
            if avishan_raise_400:
                from .exceptions import ErrorMessageException

                raise ErrorMessageException("Chosen " + cls.__name__ + " doesnt exist")  # todo: esme farsi bara model
            if avishan_raise_exception:
                from .utils.data_functions import save_traceback
                save_traceback()
                raise e
            return None

    @classmethod
    def filter(cls, avishan_to_dict: bool = False, **kwargs) -> QuerySet(
        'AvishanModel'):
        """
        [X] if empty kwargs return all
        """
        if avishan_to_dict:
            return [item.to_dict(compact=True) for item in
                    cls.filter(avishan_to_dict=False, **kwargs)]

        if len(kwargs.items()) > 0:
            return cls.objects.filter(**kwargs)
        else:
            return cls.objects.all()

    @classmethod
    def all(cls, avishan_to_dict: bool = False) -> QuerySet('AvishanModel'):
        return cls.filter(avishan_to_dict)

    @classmethod
    def create(cls, avishan_object_trace: str = "",
               avishan_prevent_clean: bool = False,
               **kwargs) -> 'AvishanModel':
        from avishan_wrapper import current_request
        """
        [O] if value error, raise true exception. translate value error
        [O] control field choices
        """

        if not avishan_prevent_clean:
            create_kwargs, after_create_kwargs = cls.__clean_input_data_for_model(kwargs, avishan_object_trace
                                                                                  )
        else:
            create_kwargs = kwargs
            after_create_kwargs = {}

        created = cls.__create(**create_kwargs)

        if after_create_kwargs:
            for key, value in after_create_kwargs.items():
                for item in value:
                    created.__getattribute__(key).add(item)
        created.save()
        return created

    def update(self, avishan_object_trace: str = "", avishan_prevent_clean: bool = False, **kwargs) -> 'AvishanModel':
        if not avishan_prevent_clean:
            in_time_kwargs, many_to_many_kwargs = self.__class__.__clean_input_data_for_model(kwargs,
                                                                                              avishan_object_trace)
        else:
            in_time_kwargs = kwargs
            many_to_many_kwargs = {}
        # todo: remove identifier attributes to prevent update on them
        # todo: check for change. if not changed, dont update
        for key, value in in_time_kwargs.items():
            # todo check values
            self.__setattr__(key, value)

        for key, value in many_to_many_kwargs.items():
            self.__getattribute__(key).clear()
            for item in value:
                self.__getattribute__(key).add(item)
        self.save()
        return self

    def remove(self) -> dict:
        # todo check access
        temp = self.to_dict()
        self.delete()
        return temp

    @classmethod
    def get_required_fields(cls) -> List[Field]:
        return [field for field in cls.get_fields() if cls.is_field_required(field)]

    @classmethod
    def define_object_action(cls, input_dict: dict):

        create_flag = get_flag = False
        for field in cls.get_full_fields():
            if cls.is_field_identifier_for_model(field) and field.name in input_dict.keys():
                get_flag = True

            elif field.name in input_dict.keys():
                create_flag = True

        if create_flag:
            if get_flag:
                return cls.ObjectAction.UPDATE
            return cls.ObjectAction.CREATE
        return cls.ObjectAction.GET

    @classmethod
    def __clean_input_data_for_model(cls, input_dict: dict, previous_object_trace: str) -> Tuple[
        dict, dict]:
        from django.db.models import DateField, DateTimeField, TimeField, OneToOneField, ForeignKey, ManyToManyField
        from .exceptions import ErrorMessageException

        object_trace = previous_object_trace + cls.class_snake_case_name()
        create_kwargs = {}
        after_create_kwargs = {}
        for field in cls.get_full_fields():
            if cls.is_field_identifier_for_model(field, on_create=True) \
                    or (isinstance(field, (DateField, TimeField)) and (field.auto_now or field.auto_now_add)):
                continue
            elif cls.is_field_required(field) and field.name not in input_dict.keys():
                raise ErrorMessageException(
                    f'Field {field.name} not found in object {cls.class_name()}, and it\'s required.'
                )

            elif field.name not in input_dict.keys():
                continue

            elif isinstance(field, (OneToOneField, ForeignKey)):
                if isinstance(input_dict[field.name], models.Model):
                    create_kwargs[field.name] = input_dict[field.name]
                else:
                    length = len(object_trace)
                    object_trace += "->" + field.name + ":"
                    create_kwargs[field.name] = field.related_model.__get_object_from_dict(input_dict[field.name],
                                                                                           object_trace)
                    object_trace = object_trace[:length]

            elif isinstance(field, ManyToManyField):

                after_create_kwargs[field.name] = []
                length = len(object_trace)
                object_trace += "->" + field.name + ":"
                for input_item in input_dict[field.name]:
                    if isinstance(input_item, models.Model):
                        item_object = input_item
                    else:
                        item_object = field.related_model.__get_object_from_dict(input_item, object_trace)
                    after_create_kwargs[field.name].append(
                        item_object
                    )
                object_trace = object_trace[:length]

            else:
                create_kwargs[field.name] = input_dict[field.name]
                if isinstance(input_dict[field.name], BchDatetime):
                    if isinstance(field, DateTimeField):
                        create_kwargs[field.name] = input_dict[field.name].to_datetime()
                    elif isinstance(field, DateField):
                        create_kwargs[field.name] = input_dict[field.name].to_date()
                    elif isinstance(field, TimeField):
                        create_kwargs[field.name] = input_dict[field.name].to_time()
                    else:
                        raise ErrorMessageException('خطای نامشخص در سیستم زمانی')
        return create_kwargs, after_create_kwargs

    @classmethod
    def __get_object_from_dict(cls, input_dict: dict, previous_object_trace: str, reach_to_object: bool = True) -> \
            'AvishanModel':
        from .exceptions import AuthException, ErrorMessageException
        from avishan_wrapper import current_request
        from .utils import status
        object_trace = previous_object_trace + cls.class_snake_case_name()

        defined_action = cls.define_object_action(input_dict)

        if defined_action == cls.ObjectAction.GET:
            # todo check access age raise kard chi aghaye mir salim?
            try:
                return cls.get(avishan_raise_exception=True,
                               object_trace=object_trace, **input_dict)
            except cls.DoesNotExist:
                current_request['response']['object_trace'] = object_trace
                current_request['response']['available_data'] = input_dict
                from .utils.data_functions import save_traceback
                save_traceback()
                raise ErrorMessageException('Object not found with available data.',
                                            status_code=status.HTTP_406_NOT_ACCEPTABLE)

        elif defined_action == cls.ObjectAction.CREATE:
            return cls.create(**input_dict)

        elif defined_action == cls.ObjectAction.UPDATE:
            try:
                found = cls.get(avishan_raise_exception=True, object_trace=object_trace, **input_dict)
            except cls.DoesNotExist:
                current_request['response']['object_trace'] = object_trace
                current_request['response']['available_data'] = input_dict
                from .utils.data_functions import save_traceback
                save_traceback()
                raise ErrorMessageException('Object not found with available data.',
                                            status_code=status.HTTP_406_NOT_ACCEPTABLE)
            return found.update(avishan_object_trace=object_trace,
                                **input_dict)
        # raise FuckException todo create fuck exception

    def to_dict(self: 'AvishanModel', compact: bool = False, except_list: list = None,
                visible_list: list = None) -> dict:
        from datetime import time
        from .utils.model_functions import filter_added_properties, filter_private_fields, filter_compact_fields, \
            filter_except_list
        # todo objectifire kamel
        # todo __dict__ ?

        dicted = {}

        if not visible_list:
            visible_list = []
        if not except_list:
            except_list = []

        field_names = []
        for field in self.get_full_fields():
            field_names.append(field.name)

        field_names = filter_added_properties(field_names, self)
        field_names = filter_private_fields(field_names, self, visible_list)
        if compact:
            field_names = filter_compact_fields(field_names, self)
        field_names = filter_except_list(field_names, except_list)

        for field in self.get_full_fields():
            if field.name not in field_names:
                continue

            value = self.__getattribute__(field.name)
            if value is None:
                dicted[field.name] = {}
            elif isinstance(field, (models.DateField)):
                dicted[field.name] = BchDatetime(value).to_dict(full=True)
            elif isinstance(field, (models.OneToOneField, models.ForeignKey)):
                dicted[field.name] = value.to_dict()
            elif isinstance(field, models.ManyToManyField):
                dicted[field.name] = []
                for item in value.all():
                    dicted[field.name].append(item.to_dict())
            elif isinstance(value, time):
                dicted[field.name] = {
                    'hour': value.hour, 'minute': value.minute, 'second': value.second, 'microsecond': value.microsecond
                }
            else:
                dicted[field.name] = value

        for field_name in filter_added_properties(field_names, self):
            if field_name in field_names and field_name not in dicted.keys():
                dicted[field_name] = self.__getattribute__(field_name)

        return dicted

    @classmethod
    def class_snake_case_name(cls) -> str:
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', cls.class_name())
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    @classmethod
    def class_name(cls) -> str:
        return cls.__name__

    @classmethod
    def class_plural_snake_case_name(cls) -> str:
        return cls.class_snake_case_name() + "s"

    @classmethod
    def app_name(cls) -> str:
        return cls._meta.app_label

    @classmethod
    def update_or_create_object(cls, fixed_kwargs: dict, new_additional_kwargs: dict) -> Tuple[
        Type['AvishanModel'], bool]:
        try:
            found = cls.get(avishan_raise_exception=True, **fixed_kwargs)
            return found.update(**{**fixed_kwargs, **new_additional_kwargs}), False
        except cls.DoesNotExist:
            return cls.create(**{**fixed_kwargs, **new_additional_kwargs}), True

    @classmethod
    def get_target_type_from_field(cls, field: Field) -> type:
        if isinstance(field, (models.CharField, models.TextField)):
            return str
        if isinstance(field, (models.IntegerField, models.AutoField)):
            return int
        if isinstance(field, models.FloatField):
            return float
        if isinstance(field, models.TimeField):
            import datetime
            return datetime.time
        if isinstance(field, models.DateField):
            import datetime
            return datetime.date
        if isinstance(field, models.DateTimeField):
            import datetime
            return datetime.datetime
        if isinstance(field, models.BooleanField):
            return bool
        if isinstance(field, models.FileField):
            return File  # todo doros she
        if isinstance(field, models.ForeignKey):
            return field.related_model

        raise NotImplementedError(
            f'get_target_type_from_field function with cls: {cls.class_name()} and field {field.name}')

    @classmethod
    def __get(cls, **kwargs):
        return cls.objects.get(**kwargs)

    @classmethod
    def __create(cls, **kwargs) -> 'AvishanModel':
        return cls.objects.create(**kwargs)


class Image(AvishanModel):
    file = models.ImageField(blank=True)
    private_fields = ('date_created',)
    list_display = ('__str__', 'date_created')

    def __str__(self):
        return self.file.name

    def to_dict(self, **kwargs) -> dict:
        temp = {}
        temp['url'] = self.file.url
        temp['id'] = self.id
        return temp
    # todo: override delete() to fully remove from hard disk
    # todo: override create to chmod
    # todo: override create with compression util


class File(AvishanModel):
    file = models.FileField(blank=True)
    private_fields = ('date_created',)
    list_display = ('file', 'date_created')

    def __str__(self):
        return self.file.name


class UserGroup(AvishanModel):
    title = models.CharField(max_length=255, unique=True)
    token_valid_seconds = models.IntegerField(default=0)
    is_base_group = models.BooleanField(default=False)

    list_display = ('title', 'is_base_group', 'token_valid_seconds')

    # todo only one is_base_group can have

    def __str__(self):
        return self.title


class User(AvishanModel):
    phone = models.CharField(max_length=255, unique=True)
    first_name = models.CharField(max_length=255, blank=True)
    last_name = models.CharField(max_length=255, blank=True)

    is_active = models.BooleanField(default=False)
    have_profile = models.BooleanField(default=False)

    date_updated = models.DateTimeField(blank=True, null=True)

    compact_fields = ('first_name', 'last_name', 'phone')
    list_display = ('phone', '__str__', 'is_active')
    private_fields = ['id', 'have_profile', 'date_updated']

    def __str__(self):
        if self.first_name or self.last_name:
            return self.first_name + " " + self.last_name
        return self.phone

    def is_in_group(self, user_group: UserGroup) -> bool:
        try:
            UserUserGroup.get(avishan_raise_exception=True, user=self, user_group=user_group)
            return True
        except UserUserGroup.DoesNotExist:
            return False

    def add_to_user_group(self, user_group: UserGroup):
        try:
            UserUserGroup.get(
                avishan_raise_exception=True,
                user=self,
                user_group=user_group
            )
        except UserUserGroup.DoesNotExist:
            UserUserGroup.create(
                user=self,
                user_group=user_group,
            )


class UserUserGroup(AvishanModel):
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='user_user_groups')
    user_group = models.ForeignKey(UserGroup, on_delete=models.CASCADE, related_name='user_user_groups')
    is_logged_in = models.BooleanField(default=False)
    date_last_login = models.DateTimeField(blank=True, null=True)
    date_last_used = models.DateTimeField(blank=True, null=True)

    list_display = ('user', 'user_group', 'date_last_login', 'date_last_used')

    def __str__(self):
        return str(self.user) + ' - ' + str(self.user_group)


class KavenegarSMS(AvishanModel):
    STATUS_TYPES = {
        'in_queue': 1,
        'scheduled': 2,
        'sent_to_telecom': 4,
        'sent_to_telecom2': 5,
        'failed': 6,
        'delivered': 10,
        'undelivered': 11,
        'user_canceled_sms': 13,
        'user_blocked_sms': 14,
        'invalid_id': 100
    }
    receptor = models.CharField(max_length=255)
    message = models.TextField()
    template_title = models.CharField(max_length=255, null=True, blank=True)

    http_status_code = models.CharField(max_length=255,
                                        blank=True)  # todo: https://kavenegar.com/rest.html#result-general
    message_id = models.CharField(max_length=255, blank=True)
    status = models.IntegerField(default=-1, blank=True)
    sender = models.CharField(max_length=255, blank=True)
    date = models.DateTimeField(blank=True, null=True)
    cost = models.IntegerField(blank=True, default=0)

    list_display = ('receptor', 'message', 'template_title', 'http_status_code', 'status', 'cost')

    def __str__(self):
        return self.receptor

    @classmethod
    def class_plural_snake_case_name(cls) -> str:
        return 'kavenegar_smses'


class ActivationCode(AvishanModel):
    code = models.CharField(max_length=255)
    user_group = models.ForeignKey(UserGroup, on_delete=models.CASCADE, null=True, blank=True)
    kavenegar_sms = models.ForeignKey(KavenegarSMS, on_delete=models.CASCADE)

    list_display = ('kavenegar_sms', 'code')

    def __str__(self):
        return self.kavenegar_sms.receptor + ' - ' + self.code


class ExceptionRecord(AvishanModel):
    class_title = models.CharField(max_length=255)
    user = models.ForeignKey(User, on_delete=models.SET_NULL, null=True, blank=True)
    user_group = models.ForeignKey(UserGroup, on_delete=models.SET_NULL, null=True, blank=True)
    status_code = models.IntegerField()
    request_url = models.TextField()
    request_method = models.CharField(max_length=255)
    request_data = models.TextField()
    request_headers = models.TextField(null=True)
    response = models.TextField()
    traceback = models.TextField()
    exception_args = models.TextField(null=True)
    checked = models.BooleanField(default=False)

    list_display = ('class_title', 'date_created', 'get_title', 'user', 'checked')
    list_filter = ('class_title', 'user', 'request_url', 'checked')
    date_hierarchy = 'date_created'

    @property
    def get_title(self):
        # try:
        if self.exception_args:
            return self.exception_args
        return self.response
    # except:
    #     return 'UNKNOWN'
# todo: create a request copy model. to keep request full data



from importlib import import_module

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Avishan init'

    def handle(self, *args, **kwargs):
        """ todo check ha
         use tz
         static
         media
        """
        init_file = import_module('avishan.init')
        init_file.init()
        self.stdout.write(self.style.SUCCESS('init function executed successfully'))



# Generated by Django 2.2.3 on 2019-08-11 12:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0002_activationcode_user_group'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExceptionRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('class_title', models.CharField(max_length=255)),
                ('url', models.CharField(max_length=255)),
                ('http_method', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



PROJECT_NAME = ''
MONITORED_URLS_START = ['/api/v1/']
MONITORED_ANONYMOUS_URLS = ['/api/v1/auth/check_phone/', '/api/v1/auth/check_code/', ]
NOT_MONITORED_URLS = []
BASE_URLS = []

JWT_KEY = '*'

USER_RESPONSE_DICT_FIELDS = ['phone', 'first_name', 'last_name']

RECOMMEND_CODE_CHOICES = 'ABCDEFGHJKLMNPQRSTUWXYZ23456789'

SMS_CODE_LENGTH = 4
SMS_CODE_VALID_SECONDS = 4 * 60
SMS_RESEND_ALLOWED_SECONDS = 90
SMS_SIGNIN_TEMPLATE = '*'
SMS_SIGNUP_TEMPLATE = '*'
KAVENEGAR_API_TOKEN = '*'
RECOMMEND_CODE_FROM_USER = 'profile__recommend_code'
AUTO_CONVERT_TO_BCHDATETIME = True

CHABOK_ACCESS_TOKEN = ''

HTTP_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS']


# Generated by Django 2.2 on 2019-09-05 17:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0007_exceptionrecord_checked'),
    ]

    operations = [
        migrations.CreateModel(
            name='Place',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('nickname', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Room',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('room_number', models.IntegerField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterField(
            model_name='usergroup',
            name='title',
            field=models.CharField(max_length=255, unique=True),
        ),
        migrations.CreateModel(
            name='Animal',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=255)),
                ('address', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='animal', to='avishan.Room')),
                ('father', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='avishan.Animal')),
                ('favorite_places', models.ManyToManyField(related_name='favorite_for', to='avishan.Place')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import sys

from .utils.bch_datetime import BchDatetime
from .utils.data_functions import save_traceback
from .utils import status
import traceback
from avishan_wrapper import current_request


class AvishanException(Exception):
    def __init__(self, from_exception: Exception = None):
        self.user = current_request['user']
        self.user_group = current_request['user_group']
        self.request_url = current_request['request'].path
        self.request_method = current_request['request'].method
        self.request_headers = None

        for key, value in current_request['request'].META.items():
            if key.startswith('HTTP_'):
                if not self.request_headers:
                    self.request_headers = ""
                self.request_headers += f'({key[5:]}: {value}),\n'

        try:
            self.request_data = current_request['request'].data if (
                    self.request_method != 'GET' or self.request_method != 'DELETE') else {}
        except AttributeError:
            self.request_data = {}

        if not current_request['traceback']:
            save_traceback()
        self.traceback = current_request['traceback']

        self.response = current_request['response']
        if not from_exception:
            self.status_code = current_request['status_code']
            self.class_title = self.__class__.__name__
            self.exception_args = None

        else:
            self.class_title = from_exception.__class__.__name__
            self.status_code = current_request['status_code'] = status.HTTP_500_INTERNAL_SERVER_ERROR
            self.exception_args = str(from_exception.args)

            current_request['response'] = {
                'error_message': 'خطای سرور'
            }

        from .models import ExceptionRecord
        ExceptionRecord.objects.create(
            class_title=self.class_title,
            user=self.user,
            user_group=self.user_group,
            status_code=self.status_code,
            request_url=self.request_url,
            request_method=self.request_method,
            request_data=self.request_data,
            request_headers=self.request_headers,
            response=self.response,
            traceback=self.traceback,
            exception_args=self.exception_args,
        )


class AuthException(AvishanException):
    ACCOUNT_NOT_FOUND = 0
    ACCOUNT_NOT_ACTIVE = 1
    TOKEN_EXPIRED = 2
    WTF = 3
    ADMIN_PERMISSION_NEEDED = 4
    INCORRECT_PASSWORD = 5
    UNAUTHORIZED_ROLE = 6
    INCORRECT_SMS_CODE = 7
    SMS_CODE_EXPIRED = 8
    INCORRECT_TOKEN = 9
    TOKEN_ERROR = 10
    ACCESS_DENIED = 11

    def __init__(self, error_type):
        current_request['response']['error_type'] = error_type
        current_request['status_code'] = status.HTTP_403_FORBIDDEN
        self.error_type = error_type

        if error_type == 0:
            error_message = 'ACCOUNT_NOT_FOUND'
        elif error_type == 1:
            error_message = 'ACCOUNT_NOT_ACTIVE'
        elif error_type == 2:
            error_message = 'TOKEN_EXPIRED'
        elif error_type == 4:
            error_message = 'ADMIN_PERMISSION_NEEDED'
        elif error_type == 5:
            error_message = 'INCORRECT_PASSWORD'
        elif error_type == 6:
            error_message = 'UNAUTHORIZED_ROLE'
        elif error_type == 7:
            error_message = 'INCORRECT_SMS_CODE'
        elif error_type == 8:
            error_message = 'SMS_CODE_EXPIRED'
        elif error_type == 9:
            error_message = 'INCORRECT_TOKEN'
        elif error_type == 10:
            error_message = 'TOKEN_ERROR'
            current_request['status_code'] = status.HTTP_424_FAILED_DEPENDENCY
        elif error_type == 11:
            error_message = 'ACCESS_DENIED'
        # todo: in che margiye
        else:
            error_message = 'WTF'

        current_request['response']['error_message'] = error_message
        super().__init__()


class ErrorMessageException(AvishanException):

    def __init__(self, message: str = None, status_code: int = status.HTTP_400_BAD_REQUEST):

        if not message:
            message = 'خطا'
        current_request['response']['error_message'] = message
        current_request['status_code'] = status_code
        super().__init__()


class ValidatorException(AvishanException):

    def __init__(self, field_name: str):
        current_request['response']['error_in_field'] = field_name
        current_request['response']['error_message'] = '"' + field_name + '" قابل قبول نیست'
        current_request['status_code'] = status.HTTP_417_EXPECTATION_FAILED
        super().__init__()


def add_warning_message(text: str):
    current_request['response']['warning_message'] = text



import json
import pprint

from django.http import JsonResponse

from .utils.data_functions import save_traceback
from .exceptions import AvishanException
from avishan_wrapper import current_request


class AvishanDecorator(object):
    def __init__(self):

        pass

    def __call__(self, view_function):
        # todo we should have method checker
        def wrapper(*args, **kwargs):
            request = current_request['request']

            if request.method != 'GET':
                from django.http.request import RawPostDataException
                try:
                    if len(request.body) > 0:
                        request.data = json.loads(request.body.decode('utf-8'))
                        # todo: hich kodoom az key ha nabayad ba "avishan*_" shoroo beshe
                    else:
                        request.data = {}
                except RawPostDataException:
                    request.data = {}

            try:
                return view_function(*args, **kwargs)
            except Exception as e:
                save_traceback()
                current_request['exception'] = e
                return JsonResponse({}, status=567)

        return wrapper


class AvishanCalculateTime(object):
    def __init__(self):
        pass

    def __call__(self, view_function):
        from .utils.bch_datetime import BchDatetime

        def wrapper(*args, **kwargs):
            start_time = BchDatetime()

            result = view_function(*args, **kwargs)

            end_time = BchDatetime()
            current_request['execution_time'] = end_time.to_unix_timestamp('microsecond') - start_time.to_unix_timestamp(
                'microsecond')

            return result

        return wrapper


class AvishanPrintRequestBody(object):
    def __init__(self):
        pass

    def __call__(self, view_function):

        def wrapper(*args, **kwargs):
            print("++++++++++++++++++++ " + view_function.__name__ + " ++++++++++++++++++++")
            print("***** REQUEST *****")
            request = current_request['request']
            if request.method != 'GET':
                try:
                    pprint.pprint(request.data)
                except AttributeError:
                    if len(request.body) > 0:
                        request.data = json.loads(request.body.decode('utf-8'))
                    else:
                        request.data = {}
                    pprint.pprint(request.data)

            try:
                result = view_function(*args, **kwargs)
            except AvishanException as e:
                print("***** RESPONSE *****")
                print("-------------------- " + view_function.__name__ + " --------------------")
                save_traceback()
                raise e

            print("***** RESPONSE *****")
            if isinstance(result, JsonResponse):
                response = json.loads(result.content.decode('utf-8'))
                pprint.pprint(response)
            print("-------------------- " + view_function.__name__ + " --------------------")

            return result

        return wrapper

# todo: part by part execution time calculator decorator



# Generated by Django 2.2 on 2019-07-25 01:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AccessRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('title', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('file', models.FileField(blank=True, upload_to='')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('file', models.ImageField(blank=True, upload_to='')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='KavenegarSMS',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('receptor', models.CharField(max_length=255)),
                ('message', models.TextField()),
                ('template_title', models.CharField(blank=True, max_length=255, null=True)),
                ('http_status_code', models.CharField(blank=True, max_length=255)),
                ('message_id', models.CharField(blank=True, max_length=255)),
                ('status', models.IntegerField(blank=True, default=-1)),
                ('sender', models.CharField(blank=True, max_length=255)),
                ('date', models.DateTimeField(blank=True, null=True)),
                ('cost', models.IntegerField(blank=True, default=0)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('phone', models.CharField(max_length=255, unique=True)),
                ('first_name', models.CharField(blank=True, max_length=255)),
                ('last_name', models.CharField(blank=True, max_length=255)),
                ('is_active', models.BooleanField(default=False)),
                ('have_profile', models.BooleanField(default=False)),
                ('date_updated', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UserGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('title', models.CharField(max_length=255)),
                ('token_valid_seconds', models.IntegerField(default=0)),
                ('is_base_group', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UserUserGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('is_logged_in', models.BooleanField(default=False)),
                ('date_last_login', models.DateTimeField(blank=True, null=True)),
                ('date_last_used', models.DateTimeField(blank=True, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_user_groups', to='avishan.User')),
                ('user_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_user_groups', to='avishan.UserGroup')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UserAccessRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('is_allowed', models.BooleanField(default=True)),
                ('access_rule', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_access_rules', to='avishan.AccessRule')),
                ('granted_by', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='granted_user_access_rules', to='avishan.User')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_access_rules', to='avishan.User')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ActivationCode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('code', models.CharField(max_length=255)),
                ('kavenegar_sms', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='avishan.KavenegarSMS')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='accessrule',
            name='user_groups',
            field=models.ManyToManyField(related_name='access_rules', to='avishan.UserGroup'),
        ),
    ]



import datetime
import json
import random
from typing import Tuple

import requests
from django.core.paginator import Paginator
from django.db.models import QuerySet

from avishan_config import RECOMMEND_CODE_CHOICES, SMS_SIGNIN_TEMPLATE, KAVENEGAR_API_TOKEN, SMS_SIGNUP_TEMPLATE, \
    CHABOK_ACCESS_TOKEN, RECOMMEND_CODE_FROM_USER
from ..utils.bch_datetime import BchDatetime
from ..models import User, KavenegarSMS, ActivationCode


def is_number(input):
    try:
        float(input)
        return True
    except ValueError:
        pass

    try:
        import unicodedata
        unicodedata.numeric(input)
        return True
    except (TypeError, ValueError):
        pass

    return False


def has_numbers(input):
    return any(char.isdigit() for char in input)


def introduce_code_length() -> int:
    return 6


def create_introduce_code() -> str:
    while True:
        new_code = ''.join(
            random.choice(RECOMMEND_CODE_CHOICES) for _ in range(introduce_code_length()))
        try:
            # todo: mese adam
            User.objects.get(**{RECOMMEND_CODE_FROM_USER: new_code})
            continue
        except User.DoesNotExist:
            return new_code


def convert_to_fa_number(text):
    text = str(text)
    text = convert_to_en_number(text)
    array = ['۰', '۱', '۲', '۳', '۴', '۵', '۶', '۷', '۸', '۹']
    result = ''
    for i in str(text):
        if i.isdigit():
            result = result + array[int(i)]
        else:
            result = result + i

    return result


def convert_to_en_number(text):
    text = str(text)
    result = ''
    array = ['۰', '۱', '۲', '۳', '۴', '۵', '۶', '۷', '۸', '۹']
    for char in text:
        if char in array:
            result += str(array.index(char))
        else:
            result += char
    return result


def create_sms_transaction(user: User, amount: int):
    # if amount == 0:
    #     return
    #
    # amount = -amount
    # from core.models import SystemTransaction
    # last_transaction = SystemTransaction.objects.filter(is_sms=True, transaction__user=user).order_by('-date_created')
    # if len(last_transaction) > 0:
    #     last_transaction = last_transaction[0]
    #     last_transaction.transaction.amount += amount
    #     last_transaction.transaction.date_created = BchDatetime().to_datetime()
    #     last_transaction.transaction.save()
    # else:
    #     from core.models import Transaction
    #     SystemTransaction.objects.create(
    #         transaction=Transaction.objects.create(
    #             user=user, amount=amount, title='هزینه پیامک'
    #         ),
    #         is_sms=True
    #     )
    # todo
    pass


def send_template_sms(phone_number: str, template_title: str, token: str, token2: str = None,
                      token3: str = None) -> KavenegarSMS:
    from avishan_wrapper import current_request
    kavenegar_sms = KavenegarSMS.objects.create(
        receptor=phone_number,
        message=token + "|" + str(token2) + "|" + str(token3),
        template_title=template_title,
        date_created=BchDatetime().to_datetime()
    )

    if template_title == SMS_SIGNIN_TEMPLATE or template_title == SMS_SIGNUP_TEMPLATE:
        activation_code = ActivationCode.objects.create(
            code=token,
            user_group=current_request['user_group'],
            kavenegar_sms=kavenegar_sms,
            date_created=BchDatetime().to_datetime()
        )
    url = "https://api.kavenegar.com/v1/" + KAVENEGAR_API_TOKEN + "/verify/lookup.json"
    querystring = {"receptor": phone_number, "token": token, "template": template_title}
    if token2:
        querystring['token2'] = token2
    if token3:
        querystring['token3'] = token3
    response = requests.request("GET", url, data="", headers={}, params=querystring)

    data = json.loads((response.content).decode('utf8'))['entries'][0]
    kavenegar_sms.message_id = data['messageid']
    kavenegar_sms.message = data['message']
    kavenegar_sms.status = data['status']
    kavenegar_sms.sender = data['sender']
    kavenegar_sms.date = datetime.datetime.fromtimestamp(data['date'])
    kavenegar_sms.cost = data['cost']
    kavenegar_sms.http_status_code = response.status_code
    kavenegar_sms.save()

    return kavenegar_sms


def send_raw_sms(phone_number: str, message: str) -> KavenegarSMS:
    kavenegar_sms = KavenegarSMS.objects.create(
        receptor=phone_number,
        message=message,
        template_title=None
    )

    url = 'https://api.kavenegar.com/v1/' + KAVENEGAR_API_TOKEN + '/sms/send.json'
    response = requests.post(
        url,
        {'receptor': phone_number, 'message': message}
    )

    data = json.loads(response.content.decode('utf8'))['entries'][0]

    kavenegar_sms.message_id = data['messageid']
    kavenegar_sms.message = data['message']
    kavenegar_sms.status = data['status']
    kavenegar_sms.sender = str(data['sender'])
    kavenegar_sms.date = datetime.datetime.fromtimestamp(data['date'])
    kavenegar_sms.cost = data['cost']
    kavenegar_sms.http_status_code = response.status_code
    kavenegar_sms.save()

    try:
        user = User.objects.get(phone=phone_number, profile__use_sms_notification=True)
        create_sms_transaction(user, kavenegar_sms.cost // 10)
    except User.DoesNotExist:
        pass

    return kavenegar_sms


def send_chabok_push(phone_number: str, title: str, body: str):
    requests.post(
        'https://sandbox.push.adpdigital.com/api/push/notifyUser/' + phone_number
        + '?access_token=' + CHABOK_ACCESS_TOKEN,
        data={'title': title, 'body': body}
    )
    # todo check for error and raise exception


def add_query_set_to_response(query_set: QuerySet, response_dict: dict, key_name: str, request,
                              compact: bool = False, ) -> dict:
    response_dict[key_name] = []
    # if len(query_set) == 0:
    #     return response_dict
    # if len(request.search) > 0:
    #     sum_searched_query_set = query_set[0:0]
    #     for searchable_field in query_set[0].searchable_fields:
    #         sum_searched_query_set = sum_searched_query_set | query_set.filter(**{
    #             searchable_field + '__contains': request.search
    #         })
    #     sum_searched_query_set.distinct()
    #     query_set = sum_searched_query_set
    # todo: implement filter and sort & search

    paginated, count, total, have_next, have_previous = paginate(query_set, request.page, request.page_size)
    response_dict[key_name] = [item.to_dict(compact=compact) for item in paginated]

    response_dict['pagination'] = {
        'count': count,
        'total': total,
        'have_next': have_next,
        'have_previous': have_previous
    }

    return response_dict


def paginate(query_set: QuerySet, page: int = 0, page_size: int = 0) -> Tuple[
    QuerySet, int, int, bool, bool]:
    """
    it returns a tuple:
    paginated query_set, pages count, total objects count, have next, have previous
    """
    if page == 0:
        return (
            query_set, 1, query_set.count(), False, False
        )
    paginated = Paginator(query_set, page_size)

    return (
        paginated.page(page).object_list, paginated.num_pages, paginated.count, paginated.page(page).has_next(),
        paginated.page(page).has_previous()
    )


def diagram_js(days: int = 7, **kwargs) -> dict:
    # todo: set a rule for colors, unlimited
    dataset = {
        'backgroundColor': 'transparent', 'label': 'sample', 'borderColor': '#000000', 'data': []
    }
    data = {
        'labels': [],
        'datasets': []
    }

    for key, value in kwargs.items():
        temp = {**dataset}
        temp['label'] = key
        temp['borderColor'] = "#" + str(random.randint(0, 10)) + str(random.randint(0, 10)) + str(
            random.randint(0, 10)) + str(random.randint(0, 10)) + str(random.randint(0, 10)) + str(
            random.randint(0, 10))
        temp['model'] = value['model']
        temp['filter'] = value['filter']
        temp['added_filters'] = value.get('added_filters')
        data['datasets'].append(temp)

    end_datetime = BchDatetime().to_datetime()
    start_datetime = BchDatetime().load_time(datetime.time(0, 0, 0, 0)).to_datetime()

    for i in range(days):
        for dataset in data['datasets']:
            filter_dict = {
                dataset['filter'] + "__gte": start_datetime,
                dataset['filter'] + "__lt": end_datetime
            }
            if dataset['added_filters']:
                for item in dataset['added_filters']:
                    filter_dict[item[0]] = item[1]
            dataset['data'].append(
                dataset['model'].objects.filter(**filter_dict).count()
            )
        end_datetime = BchDatetime(start_datetime).to_datetime()
        start_datetime = start_datetime - datetime.timedelta(days=1)

    data['labels'].reverse()
    for dataset in data['datasets']:
        dataset['data'].reverse()
        del dataset['model']
    return data


# todo: yebar begirim ba filter, baad tooye for age lazem darim check konim hast behtare ya harbar get bezanim too for?


def save_traceback():
    import sys, traceback
    from avishan_wrapper import current_request
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = traceback.TracebackException(
        exc_type, exc_value, exc_tb,
    )
    current_request['traceback'] = ''.join(tbe.format())



import datetime

from ..exceptions import ValidatorException
from ..utils.data_functions import convert_to_en_number, convert_to_fa_number, introduce_code_length, has_numbers

# todo in minimum length ha max nadarn?


def validate_phone_number(input: str, country_code: str = '98', phone_start_number: str = '09') -> str:
    input = convert_to_en_number(input)
    input = input.replace(" ", "")
    input = input.replace("-", "")

    if input.startswith("00"):
        if not input.startswith("00" + country_code):
            raise ValidatorException('شماره موبایل')
        if input.startswith("00" + country_code + phone_start_number):
            input = "00" + country_code + input[5:]
    elif input.startswith("+"):
        if not input.startswith("+" + country_code):
            raise ValidatorException('شماره موبایل')
        input = "00" + input[1:]
        if input.startswith("00" + country_code + phone_start_number):
            input = "00" + country_code + input[5:]
    elif input.startswith(phone_start_number):
        input = "00" + country_code + input[1:]

    if len(input) != 14 or not input.isdigit():
        raise ValidatorException('شماره موبایل')

    return input


def validate_text(input: str, blank: bool = True) -> str:
    input = input.strip()
    input = convert_to_fa_number(input)

    if not blank and len(input) == 0:
        raise ValidatorException('متن')

    return input


def validate_recommend_code(input: str) -> str:
    input = validate_text(input)

    input = convert_to_en_number(input)

    input = input.upper()

    return input


def validate_first_name(input):
    input = input.strip()

    if has_numbers(input) or len(input) < 2:
        raise ValidatorException('نام')

    return input


def validate_last_name(input):
    input = input.strip()

    if has_numbers(input) or len(input) < 2:
        raise ValidatorException('نام خانوادگی')

    return input


def validate_ferdowsi_student_id(input):
    input = validate_text(input, blank=False)

    if not input.isdigit():
        raise ValidatorException('شماره دانشجویی')
    return input


def validate_plate(plate_a, plate_b, plate_c, plate_d):
    plate_a = validate_text(convert_to_fa_number(plate_a), blank=False)
    plate_b = validate_text(convert_to_fa_number(plate_b), blank=False)
    plate_c = validate_text(convert_to_fa_number(plate_c), blank=False)
    plate_d = validate_text(convert_to_fa_number(plate_d), blank=False)

    if plate_b not in ['ب', 'ج', 'د', 'س', 'ص', 'ط', 'ق', 'ل', 'م', 'ن', 'و', 'ه', 'ی', 'الف', 'پ', 'ت', 'ث', 'ز', 'ژ',
                       'ش', 'ع', 'ف', 'ک', 'گ', 'D', 'S', 'd', 's', 'ي']:
        raise ValidatorException('پلاک')

    if not plate_a.isdigit() or not plate_c.isdigit() or not plate_d.isdigit():
        raise ValidatorException('پلاک')

    return plate_a, plate_b, plate_c, plate_d


def validate_time(input: dict, name: str) -> datetime.time:
    return datetime.time(int(input['hour']), int(input['minute']))



# Generated by Django 2.2 on 2019-09-02 18:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0005_auto_20190818_0336'),
    ]

    operations = [
        migrations.AddField(
            model_name='exceptionrecord',
            name='exception_args',
            field=models.TextField(null=True),
        ),
        migrations.AddField(
            model_name='exceptionrecord',
            name='request_headers',
            field=models.TextField(null=True),
        ),
    ]



from django.contrib import admin
from django.contrib.admin.sites import AlreadyRegistered

from .utils.model_functions import find_non_abstract_models

# todo auto add classes for admin
for model in find_non_abstract_models():
    model_admin_dict = {
        'date_hierarchy': model.date_hierarchy,
        'list_filter': model.list_filter,
        'list_max_show_all': model.list_max_show_all,
        'list_per_page': model.list_per_page,
        'raw_id_fields': model.raw_id_fields,
        'readonly_fields': model.readonly_fields,
        'search_fields': model.search_fields,
    }
    if model.list_display:
        model_admin_dict['list_display'] = model.list_display
    model_admin = type(model.class_snake_case_name() + "_admin", (admin.ModelAdmin,), model_admin_dict)
    try:
        admin.site.register(model, model_admin)
    except AlreadyRegistered:
        pass



from django.core.management.base import BaseCommand


# todo: these arent init. these are configs
# todo: create some functions to help writing configs like have_model & other
class Command(BaseCommand):
    help = 'Execute init commands from <app>/inti.py init function'

    def add_arguments(self, parser):
        parser.add_argument('app_name', type=str, help='app name')

    def handle(self, *args, **kwargs):
        from importlib import import_module
        app_name = kwargs['app_name']
        try:
            import_module(app_name)
        except ModuleNotFoundError:
            self.stdout.write(self.style.ERROR('%s package not found' % app_name))
            return
        try:
            init_file = import_module(app_name + ".init")
        except ModuleNotFoundError:
            self.stdout.write(self.style.ERROR('%s/init.py file not found' % app_name))
            return
        try:
            init_file.init()
            self.stdout.write(self.style.SUCCESS('init function executed successfully'))
        except AttributeError as e:
            self.stdout.write(self.style.ERROR(e.args))
            self.stdout.write(self.style.ERROR('init function not located in %s/init.py' % app_name))




# Generated by Django 2.2 on 2019-09-03 11:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0006_auto_20190902_1801'),
    ]

    operations = [
        migrations.AddField(
            model_name='exceptionrecord',
            name='checked',
            field=models.BooleanField(default=False),
        ),
    ]



from django.urls import path
from .views import *

urlpatterns = [
    path('auth/check_phone/<str:role>/<str:phone>/', check_user_phone),
    path('auth/check_code/<str:role>/<str:phone>/<str:code>/', check_user_code),

    path('base/<str:model_plural_name>/', avishan_model_store),
    path('base/<str:model_plural_name>/<int:id>/', avishan_model_detail),
]



from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist


def init():
    pass
    # try:
    #     get_user_model().objects.get(username='afshari9978')
    # except ObjectDoesNotExist:
    #     get_user_model()._default_manager.db_manager('default').create_superuser(**{
    #         'username': 'afshari9978',
    #         'email': 'afshari9978@gmail.com',
    #         'password': '',
    #         'first_name': 'Morteza',
    #         'last_name': 'Afshari'
    #     })



# Generated by Django 2.2.3 on 2019-08-14 16:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0003_exceptionrecord'),
    ]

    operations = [
        migrations.AddField(
            model_name='accessrule',
            name='crud_type',
            field=models.CharField(blank=True, default=None, max_length=255, null=True),
        ),
    ]



import jwt

from .data_functions import save_traceback
from avishan_config import JWT_KEY
from ..exceptions import AuthException
from .bch_datetime import BchDatetime
from ..models import UserGroup, UserUserGroup, User


def sign_in_user(user, user_group):
    user_user_group = verify_user(user, user_group)
    user_user_group.is_logged_in = True
    user_user_group.date_last_login = BchDatetime().to_datetime()
    user.save()
    from avishan_wrapper import current_request
    current_request['response']['token'] = encode_token(user_user_group)
    current_request['response']['user'] = user.to_dict(compact=True)


def verify_user(user_object: User = None, try_on_group: UserGroup = None) -> UserUserGroup:
    if not user_object and not try_on_group:
        from avishan_wrapper import current_request
        user_object = current_request['user']
        try_on_group = current_request['user_group']

    if not user_object.is_active:
        raise AuthException(AuthException.ACCOUNT_NOT_ACTIVE)
    try:
        return UserUserGroup.get(
            avishan_raise_exception=True, user=user_object, user_group=try_on_group
        )
    except UserUserGroup.DoesNotExist:
        save_traceback()
        raise AuthException(AuthException.UNAUTHORIZED_ROLE)


def encode_token(user_user_group: UserUserGroup) -> str:
    now = BchDatetime()
    user_user_group.date_last_used = now.to_datetime()
    user_user_group.save()
    token_data = {
        'created_at': now.to_unix_timestamp(),
        'exp': (now + user_user_group.user_group.token_valid_seconds).to_unix_timestamp(),
        'user_user_group_id': user_user_group.id
    }
    return jwt.encode(token_data,
                      JWT_KEY,
                      algorithm='HS256'
                      ).decode("utf-8")


def decode_token(token: str) -> dict:
    if not token:
        print("token: " + token)
        raise AuthException(AuthException.INCORRECT_TOKEN)
    try:
        return jwt.decode(token, JWT_KEY, algorithms=['HS256'])
    except jwt.exceptions.ExpiredSignatureError:
        save_traceback()
        raise AuthException(AuthException.TOKEN_EXPIRED)
    except:
        print("token: " + token)
        save_traceback()
        raise AuthException(AuthException.TOKEN_ERROR)


def generate_token(decoded_token, token, user: User, user_group: UserGroup) -> str:
    # todo is it efficient to create a new token
    return encode_token(UserUserGroup.objects.get(user=user, user_group=user_group))



from typing import TextIO, Type

from django.db.models import Field

from ..models import AvishanModel
from .model_functions import find_models, get_app_names


def create_raml_documentation(project_name: str):
    raml = {}
    write_header(raml, project_name)
    write_security(raml)
    write_types(raml)
    write_base_resources(raml)

    write_raml_to_file(raml, project_name)


def open_app_documentation_file(project_name: str) -> TextIO:
    return open('%s_documentation.raml' % project_name, 'w')


def write_raml_to_file(raml: dict, project_name: str):
    file = open_app_documentation_file(project_name)
    file.writelines([
        '#%RAML 1.0\n',
        'title: %s\n' % raml['title'],
        'version: %s\n' % raml['version'],
        'baseUri: %s\n' % raml['baseUri'],
    ])
    write_dict_to_file(raml['types'], file, 'types', 0)
    # todo save dict to file
    file.close()


def write_header(raml: dict, project_name: str):
    from avishan_config import BASE_URLS
    raml['title'] = project_name
    raml['version'] = 'v1'
    raml['baseUri'] = BASE_URLS[0]


def write_types(raml: dict):
    raml['types'] = {}
    for model in find_models('avishan'):
        raml['types'][model.__name__] = describe_model(model)
    for app_name in get_app_names():
        for model in find_models(app_name):
            raml['types'][model.__name__] = describe_model(model)
    if raml['types'] == {}:
        del raml['types']


def describe_model(model: Type[AvishanModel]) -> dict:
    model_dict = {
        'type': 'object',
        'properties': {}
    }

    for field in model.get_fields():
        model_dict['properties'][field.name] = describe_field(field)

    temp = describe_model_examples(model)
    if temp != {}:
        model_dict['examples'] = temp
    return model_dict


def describe_field(field: Field) -> dict:
    field_dict = {}

    from django.db.models import DateTimeField, DateField, TimeField, CharField, TextField, FileField, \
        BooleanField, IntegerField, FloatField, ManyToManyField, ForeignKey, AutoField
    if isinstance(field, (CharField, TextField)):
        field_dict['type'] = "string"
    elif isinstance(field, (IntegerField, FloatField, AutoField)):
        field_dict['type'] = 'number'
    elif isinstance(field, TimeField):
        field_dict['type'] = 'time-only'
    elif isinstance(field, DateTimeField):
        field_dict['type'] = 'datetime'
    elif isinstance(field, DateField):
        field_dict['type'] = 'date-only'
    elif isinstance(field, BooleanField):
        field_dict['type'] = 'boolean'
    elif isinstance(field, FileField):
        field_dict['type'] = 'file'
    elif isinstance(field, ManyToManyField):
        field_dict['type'] = field.related_model.class_name()
    elif isinstance(field, ForeignKey):
        field_dict['type'] = field.related_model.class_name()

    field_dict['required'] = AvishanModel.is_field_required(field)
    return field_dict


def describe_model_examples(model: Type[AvishanModel]) -> dict:
    # todo
    return {}


def write_model_additional_data(raml: dict, model: Type[AvishanModel]):
    # todo
    pass


def get_indent(stage: int) -> str:
    return '  ' * stage


def write_dict_to_file(data: dict, file: TextIO, name: str, indent_stage: int):
    file.write(get_indent(indent_stage) + name + ":\n")
    for key, value in data.items():
        if isinstance(value, dict):
            write_dict_to_file(value, file, key, indent_stage + 1)
        else:
            if value is True:
                value = 'true'
            elif value is False:
                value = 'false'
            file.write(get_indent(indent_stage + 1) + "%s: %s\n" % (key, value))


def write_base_resources(raml: dict):
    raml['/api'] = {
        '/v1': {
            'description': 'the first version of api',
            '/test': {
                'description': 'its just a test'
            },
            '/auth': {
                'description': 'authentication part of project',
                '/check_phone/{role}/{phone}': {
                    'get': {
                        'uriParameters': {
                            'role': {
                                'type': 'string',
                                'description': 'can be chosen from: user, admin'
                            },
                            'phone': {
                                'type': 'string',
                                'description': 'user entered phone'
                            },
                        },
                        'responses': {
                            '200': {
                                'body': {
                                    'user': 'User',
                                    'token': 'string'
                                },
                                'description': 'old user'
                            },
                            '201': {
                                'body': {
                                    'user': 'User',
                                    'token': 'string'
                                },
                                'description': 'new user'
                            },
                            '408': {
                                'description': 'پیامک قبلی همین چند دقیقه پیش اومده'
                            },
                            '400': {
                                'description': 'global errors',
                                'examples': {
                                    '1': 'حساب کاربری پیدا نشد'
                                }
                            }
                        }
                    },
                    '/check_code/{role}/{phone}/{code}': {
                        'get': {
                            'uriParameters': {
                                'role': {
                                    'type': 'string',
                                    'description': 'can be chosen from: user, admin'
                                },
                                'phone': {
                                    'type': 'string',
                                    'description': 'user entered phone'
                                },
                                'code': {
                                    'type': 'string',
                                    'description': 'user entered code'
                                }
                            },
                            'responses': {
                                '200': {
                                    'body': {
                                        'user': 'User',
                                        'token': 'string'
                                    },
                                    'description': 'old user'
                                },
                                '201': {
                                    'body': {
                                        'user': 'User',
                                        'token': 'string'
                                    },
                                    'description': 'new user'
                                },
                                '400': {
                                    'description': 'global errors',
                                    'examples': {
                                        '1': 'کد ورود برای این شماره پیدا نشد، لطفا مجددا کد دریافت کنید',
                                        '2': 'زمان اعتبار این کد به اتمام رسیده، لطفا مجددا کد دریافت کنید',
                                        '3': 'کد وارد شده اشتباه است',
                                        '4': 'حساب کاربری پیدا نشد'
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    raml['/api'] = {

    }


def write_security(raml: dict):
    raml['securitySchemes'] = {
        'token-authentication': {
            'type': 'Pass Through',
            'describedBy': {
                'headers': {
                    'token': {
                        'type': 'string'
                    }
                }
            }
        }
    }



import random

from django.http import JsonResponse

from .utils.model_functions import find_models, find_model_by_plural_name
from .decorators import AvishanDecorator
from .exceptions import ErrorMessageException, AuthException
from .utils.validator_functions import validate_phone_number
from avishan_config import SMS_CODE_LENGTH, SMS_RESEND_ALLOWED_SECONDS, SMS_SIGNIN_TEMPLATE, SMS_SIGNUP_TEMPLATE, \
    SMS_CODE_VALID_SECONDS
from .models import *
from .utils.bch_datetime import BchDatetime
from .utils import status
from .utils.auth_functions import verify_user, sign_in_user
from .utils.data_functions import send_template_sms


@AvishanDecorator()
def check_user_phone(request, role, phone):
    from avishan_wrapper import current_request
    phone = validate_phone_number(phone)
    current_request['user_group'] = UserGroup.get(title=role)
    if not current_request['user_group']:
        raise AuthException(AuthException.UNAUTHORIZED_ROLE)

    # todo kevenegar takes too long
    # todo email and username and etc...
    current_request['response']['phone'] = phone
    code = str(random.randrange(10 ** (SMS_CODE_LENGTH - 1), 10 ** SMS_CODE_LENGTH - 1))
    try:
        old_activation_code = ActivationCode.objects.get(kavenegar_sms__receptor=phone,
                                                         user_group=current_request['user_group'])
    except ActivationCode.DoesNotExist:
        old_activation_code = None

    if old_activation_code and (
            BchDatetime() - BchDatetime(old_activation_code.date_created)).total_seconds() < SMS_RESEND_ALLOWED_SECONDS:
        raise ErrorMessageException('برای ارسال مجدد پیامک، کمی بعد تلاش کنید',
                                    status_code=status.HTTP_408_REQUEST_TIMEOUT)
    if old_activation_code:
        old_activation_code.delete()

    try:
        user = User.objects.get(phone=phone)
        verify_user(user, current_request['user_group'])

        send_template_sms(phone, SMS_SIGNIN_TEMPLATE, code)

    except User.DoesNotExist:
        user = None
        if not current_request['user_group'].is_base_group:
            raise ErrorMessageException('حساب کاربری پیدا نشد')

        send_template_sms(phone, SMS_SIGNUP_TEMPLATE, code)

    if user and user.have_profile:
        return JsonResponse(current_request['response'])
    return JsonResponse(current_request['response'], status=status.HTTP_201_CREATED)


@AvishanDecorator()
def check_user_code(request, role, phone, code):
    from avishan_wrapper import current_request
    phone = validate_phone_number(phone)
    current_request['response']['phone'] = phone
    current_request['user_group'] = UserGroup.get(title=role)
    if not current_request['user_group']:
        raise AuthException(AuthException.WTF)

    try:
        activation_code = ActivationCode.objects.get(kavenegar_sms__receptor=phone,
                                                     user_group=current_request['user_group'])
    except ActivationCode.DoesNotExist:
        raise ErrorMessageException('کد ورود برای این شماره پیدا نشد، لطفا مجددا کد دریافت کنید')

    if (BchDatetime() - BchDatetime(activation_code.date_created)).total_seconds() > SMS_CODE_VALID_SECONDS:
        activation_code.delete()
        raise ErrorMessageException('زمان اعتبار این کد به اتمام رسیده، لطفا مجددا کد دریافت کنید')

    if activation_code.code != code:
        raise ErrorMessageException('کد وارد شده اشتباه است')

    created = False
    try:
        user = User.get(avishan_raise_exception=True, phone=phone)
        current_request['user'] = user
    except User.DoesNotExist:
        if not current_request['user_group'].is_base_group:
            raise ErrorMessageException('حساب کاربری پیدا نشد')

        # todo har user natoone be grooh haye dige add kone
        user = User.objects.create(
            phone=phone,
            is_active=True,
        )
        current_request['user'] = user
        created = True
        user.add_to_user_group(current_request['user_group'])

    activation_code.delete()
    sign_in_user(current_request['user'], current_request['user_group'])

    if created or not user.have_profile:
        return JsonResponse(current_request['response'], status=status.HTTP_201_CREATED)
    return JsonResponse(current_request['response'])


@AvishanDecorator()
def avishan_model_store(request, model_plural_name):
    from avishan_wrapper import current_request

    model = find_model_by_plural_name(model_plural_name)
    if not model:
        raise ErrorMessageException(f'AvishanError: Model not found')

    if request.method == 'GET':
        current_request['response'][model.class_plural_snake_case_name()] = [item.to_dict() for item in model.all()]

    elif request.method == 'POST':
        current_request['response'][model.class_snake_case_name()] = model.create(
            **request.data[model.class_snake_case_name()]
        ).to_dict()

    return JsonResponse(current_request['response'])


@AvishanDecorator()
def avishan_model_detail(request, model_plural_name, id):
    from avishan_wrapper import current_request
    # todo: check if model available, else raise error

    model = find_model_by_plural_name(model_plural_name)
    if not model:
        raise ErrorMessageException('AvishanError: Model not found')

    record = model.get(avishan_raise_400=True, id=id)

    if request.method == 'GET':
        current_request['response'][model.class_snake_case_name()] = record.to_dict()

    elif request.method == 'PUT':
        current_request['response'][model.class_snake_case_name()] = record.update(
            **request.data[model.class_snake_case_name()]).to_dict()

    elif request.method == 'DELETE':
        current_request['response'][model.class_snake_case_name()] = record.remove()

    return JsonResponse(current_request['response'])



# Generated by Django 2.2.3 on 2019-09-13 18:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0009_auto_20190911_1929'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='useraccessrule',
            name='access_rule',
        ),
        migrations.RemoveField(
            model_name='useraccessrule',
            name='granted_by',
        ),
        migrations.RemoveField(
            model_name='useraccessrule',
            name='user',
        ),
        migrations.DeleteModel(
            name='AccessRule',
        ),
        migrations.DeleteModel(
            name='UserAccessRule',
        ),
    ]



# Generated by Django 2.2 on 2019-09-11 19:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0008_auto_20190905_1748'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Animal',
        ),
        migrations.DeleteModel(
            name='Place',
        ),
        migrations.DeleteModel(
            name='Room',
        ),
    ]



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup(
    name='django-avishan',
    version='0.1.11',
    packages=find_packages(),
    description='Avishan is a set of tools for building fast, comfortable and reliable django apps',
    long_description=README,
    author='Morteza Afshari',
    author_email='afshari9978@gmail.com',
    url='https://gitlab.com/afshari9978/avishan',
    license='MIT',
    install_requires=[
        'Django>=2.2.5',
        'avishan_wrapper==0.1.0',
        'django-cors-headers==2.4.0',
        'gunicorn==19.9.0',
        'kavenegar==1.1.2',
        'Khayyam==3.0.17',
        'Pillow==6.0.0',
        'PyJWT==1.7.0'
    ]
)


# Generated by Django 2.2.5 on 2019-09-15 10:44

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0010_auto_20190913_1817'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='exceptionrecord',
            name='datetime',
        ),
    ]



import importlib
import pkgutil
from datetime import datetime, date, time
from typing import Union, List, Optional, Type

from django.db.models import QuerySet, Model
from django.db.models.fields.files import ImageFieldFile, FieldFile

from ..models import AvishanModel, Image
from ..utils.bch_datetime import BchDatetime


# def objectify_model(model: Union[Model, QuerySet, list], compact: bool = False, except_list: list = None,
#                     visible_list: list = None) -> Union[dict, List[dict]]:
#     if isinstance(model, QuerySet):
#         objectified = []
#         for item in model:
#             objectified.append(
#                 objectify_model(item, compact=compact, except_list=except_list, visible_list=visible_list))
#         return objectified
#
#     objectified = {}
#
#     if model is None:
#         return objectified
#
#     if not visible_list:
#         visible_list = []
#     if not except_list:
#         except_list = []
#     field_names = []
#     for field in model._meta.fields:
#         field_names.append(field.name)
#
#     field_names = filter_added_properties(field_names, model)
#     field_names = filter_private_fields(field_names, model, visible_list)
#     if compact:
#         field_names = filter_compact_fields(field_names, model)
#     field_names = filter_except_list(field_names, except_list)
#
#     for field_name in field_names:
#         data = model.__getattribute__(field_name)
#
#         # if not data and ('date' in field_name or 'time' in field_name):
#         #     objectified[field_name] = {}
#
#         if isinstance(data, datetime) or isinstance(data, date):
#             objectified[field_name] = BchDatetime(data).to_dict(full=True)
#         elif isinstance(data, Model):
#             objectified[field_name] = objectify_model(data)
#         elif isinstance(data, ImageFieldFile) or isinstance(data, FieldFile):
#             objectified['url'] = data.url
#         elif isinstance(data, time):
#             objectified[field_name] = {
#                 'hour': data.hour, 'minute': data.minute, 'second': data.second, 'microsecond': data.microsecond
#             }
#         else:
#             objectified[field_name] = data
#
#     return objectified


def filter_added_properties(field_names, model: Model):
    new_field_names = field_names[:]
    try:
        added_properties = model.added_properties
    except AttributeError:
        added_properties = []

    for added_property in added_properties:
        if added_property not in new_field_names:
            new_field_names.append(added_property)

    return new_field_names


def filter_private_fields(field_names: list, model: Model, visible_list: list):
    try:
        private_fields = model.private_fields
    except AttributeError:
        private_fields = []

    if len(visible_list) > 0:
        for field_name in field_names[:]:
            if field_name not in visible_list:
                field_names.remove(field_name)
    else:
        for field_name in field_names[:]:
            if field_name in private_fields:
                field_names.remove(field_name)

    return field_names


def filter_compact_fields(field_names, model: Model):
    try:
        compact_fields = model.compact_fields
    except AttributeError:
        compact_fields = field_names

    for field_name in field_names[:]:
        if field_name not in compact_fields:
            field_names.remove(field_name)

    return field_names


def filter_except_list(field_names, except_list):
    for field_name in field_names[:]:
        if field_name in except_list:
            field_names.remove(field_name)

    return field_names


def find_models(app_name: str = None) -> List[Type[AvishanModel]]:
    if not app_name:
        total = AvishanModel.__subclasses__()
        for item in total[:]:
            if len(item.__subclasses__()) > 0:
                total += item.__subclasses__() # todo should be recursive
        return list(set(total))
    return [x for x in find_models() if x._meta.app_label == app_name]


def find_non_abstract_models(app_name: str = None) -> List[Type[AvishanModel]]:
    return [x for x in find_models(app_name) if x._meta.abstract is False]


def find_model_by_plural_name(name: str) -> Optional[Type[AvishanModel]]:
    for model in find_models():
        if model.class_plural_snake_case_name() == name:
            return model
    return None


def get_sum_on_field(query_set: QuerySet, field_name: str) -> int:
    from django.db.models import Sum
    total = query_set.aggregate(Sum(field_name))[field_name + "__sum"]
    if total:
        return total
    return 0



def get_app_names() -> List[str]:
    from django.apps import apps
    return [key.name for key in apps.get_app_configs() if
            (not key.name.startswith('django.') and key.name != 'avishan')]


def save_image_from_url(url: str) -> Image:
    import requests

    from django.core.files import File
    from django.core.files.base import ContentFile
    from django.core.files.temp import NamedTemporaryFile

    # r = requests.get(url)
    name = url.split('/')[-1]
    if '.' not in name:
        name = url.split('/')[-2]

    image_temp = open(url, 'rb')
    # img_temp = NamedTemporaryFile(delete=True)
    # img_temp.write(r.content)
    # img_temp.flush()

    image = Image.objects.create()
    image.file.save(name, File(image_temp), save=True)

    image.save()
    return image



from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'help'

    def add_arguments(self, parser):
        pass
        # parser.add_argument('name', type=int, help='help') # arg
        # parser.add_argument('--name', type=int, help='help')  # optional arg
        # parser.add_argument('-a', '--admin', action='store_true', help='Create an admin account') # flag
        # parser.add_argument('user_id', nargs='+', type=int, help='User ID') # list of args

    def handle(self, *args, **kwargs):
        pass

    """
    styles: 
    self.style.ERROR
    self.style.NOTICE
    self.style.SUCCESS
    self.style.WARNING
    
    message = self._get_input_message(field)
    input_value = self.get_input_data(field, message)
    
    """



# Generated by Django 2.2.3 on 2019-08-02 15:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='activationcode',
            name='user_group',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='avishan.UserGroup'),
        ),
    ]
